"""
Contains classes for parsing DLPOLY topology and parameter files
"""
__all__ = ['DlpolyFieldFile', 'DlpolyConfigFile']

from ..dlpoly.dlpolyfield import DlpolyFieldFile
from ..dlpoly.dlpolyconfig import DlpolyConfigFile
